from contextvars import ContextVar
from contextlib import contextmanager

from graia.application.entities import UploadMethods

application = ContextVar("application")
event = ContextVar("event")
event_loop = ContextVar("event_loop")
broadcast = ContextVar("broadcast")

# for image
# sendGroupMessage 等发送message的指令将set该上下文条目.
image_method = ContextVar("image_method")

@contextmanager
def enter_message_send_context(method: UploadMethods):
    t = image_method.set(method)
    yield
    image_method.reset(t)

@contextmanager
def enter_context(app=None, event_i=None):
    t1 = None
    t2 = None
    t3 = None
    t4 = None

    if app:
        t1 = application.set(app)
        t3 = event_loop.set(app.broadcast.loop)
        t4 = broadcast.set(app.broadcast)
    if event_i:
        t2 = event.set(event_i)

    yield

    if t1:
        application.reset(t1)
    if all([t2, t3, t4]):
        event.reset(t2)
        event_loop.reset(t3)
        broadcast.reset(t4)


from __future__ import annotations
from typing import (Dict, List, NoReturn, Sequence, Tuple, Type, Union)

from graia.application.exceptions import EntangledSuperposition
from graia.broadcast.utilles import run_always_await
from pydantic import BaseModel

from .elements import ExternalElement, InternalElement
import regex

T = Union[InternalElement, ExternalElement]

def raiser(error):
    raise error

class MessageChain(BaseModel):
    """即 "消息链", 被用于承载整个消息内容的数据结构, 包含有一有序列表, 包含有继承了 Element 的各式类实例.

    Examples:
    1. 你可以使用 `MessageChain.create` 方法创建一个消息链:
    ``` python
    MessageChain.create([
        Plain("这是盛放在这个消息链中的一个 Plain 元素")
    ])
    ```
    2. 你可以使用 `MessageChain.isImmutable` 方法判定消息链的可变型:
    ``` python
    print(message.isImmutable()) # 监听器获取到的消息链默认为 False.
    ```
    3. 你可以使用 `MessageChain.asMutable` 和 `MessageChain.asImmutable` 方法分别获得可变与不可变的消息链.
    4. 你可以使用 `MessageChain.isSendable` 方法检查消息链是否可以被 完整无误 的发送.
    5. 使用 `MessageChain.asSendable` 方法, 将自动过滤原消息链中的无法发送的元素, 并返回一个新的, 可被发送的消息链.
    6. `MessageChain.has` 方法可用于判断特定的元素类型是否存在于消息链中:
    ``` python
    print(message.has(At))
    # 使用 in 运算符也可以
    print(At in message)
    ```
    7. 可以使用 `MessageChain.get` 方法获取消息链中的所有特定类型的元素:
    ``` python
    print(message.get(Image)) # -> List[Image]
    # 使用类似取出列表中元素的形式也可以:
    print(message[Image]) # -> List[Image]
    ```
    8. 使用 `MessageChain.asDisplay` 方法可以获取到字符串形式表示的消息, 至于字面意思, 看示例:
    ``` python
    print(MessageChain.create([
        Plain("text"), At(123, display="某人"), Image(...)
    ]).asDisplay()) # -> "text@某人 [图片]"
    ```
    9. 使用 `MessageChain.join` 方法可以拼接多个消息链:
    ``` python
    MessageChain.join(
        message1, message2, message3, ...
    ) # -> total_message
    ```
    10. `MessageChain.plusWith` 方法将在现有的基础上将另一消息链拼接到原来实例的尾部, 并生成, 返回新的实例; 该方法不改变原有和传入的实例.
    11. `MessageChain.plus` 方法将在现有的基础上将另一消息链拼接到原来实例的尾部; 该方法更改了原有的实例, 并要求 `isMutable` 方法返回 `True` 才可以执行.
    """
    __root__: Union[List[T], Tuple[T]]

    @classmethod
    def create(cls, elements: Sequence[T]) -> "MessageChain":
        """从传入的序列(可以是元组 tuple, 也可以是列表 list) 创建消息链.

        Args:
            elements (Sequence[T]): 包含且仅包含消息元素的序列

        Returns:
            MessageChain: 以传入的序列作为所承载消息的消息链
        """
        return cls(__root__=elements)

    @classmethod
    def parse_obj(cls: Type['MessageChain'], obj: List[T]) -> 'MessageChain':
        """内部接口, 会自动将作为外部态的消息元素转为内部态.

        Args:
            obj (List[T]): 需要反序列化的对象

        Returns:
            MessageChain: 内部承载有尽量有效的内部态消息元素的消息链
        """
        handled_elements = []
        for i in obj:
            if isinstance(i, InternalElement):
                handled_elements.append(i)
            elif isinstance(i, ExternalElement):
                for ii in InternalElement.__subclasses__():
                    if ii.__name__ == i.__class__.__name__:
                        handled_elements.append(ii.fromExternal(i))
            elif isinstance(i, dict) and "type" in i:
                for ii in ExternalElement.__subclasses__():
                    if ii.__name__ == i['type']:
                        for iii in InternalElement.__subclasses__():
                            if iii.__name__ == i['type']:
                                handled_elements.append(iii.fromExternal(ii.parse_obj(i)))
        return cls(__root__=tuple(handled_elements)) # 默认是不可变型

    @property
    def isImmutable(self) -> bool:
        """判断消息链是否不可变

        Returns:
            bool: 判断结果, `True` 为不可变, `False` 为可变 
        """
        return isinstance(self.__root__, tuple)

    def asMutable(self) -> "MessageChain":
        """将消息链转换为可变形态的消息链

        Returns:
            MessageChain: 内部消息结构可变的消息链
        """
        return MessageChain(__root__=list(self.__root__))

    def asImmutable(self) -> "MessageChain":
        """将消息链转换为不可变形态的消息链

        Returns:
            MessageChain: 内部消息结构不可变的消息链
        """
        return MessageChain(__root__=tuple(self.__root__))

    @property
    def isSendable(self) -> bool:
        """判断消息链是否可以被 sendGroupMessage 等发送消息的方法正确发送的方法, 注意, 这个方法并不是万能的.

        Returns:
            bool: 判断的结果, True 为可发送, False 则反之.
        """
        return all(all([
            isinstance(i, (InternalElement, ExternalElement)),
            hasattr(i, "toExternal"),
            getattr(i.__class__, "toExternal") != InternalElement.toExternal
        ]) for i in self.__root__)

    def asSendable(self) -> "MessageChain":
        """将消息链尽量转换为能够为发送消息的方法正确发送的新消息链, 该方法不保证转换无误差性, 且不保证转化后是否可以被发送.

        Returns:
            MessageChain: 返回的可能可以正确发送的消息链.
        """
        return MessageChain(__root__=tuple([i for i in self.__root__ if all([
            isinstance(i, InternalElement),
            hasattr(i, "toExternal"),
            getattr(i.__class__, "toExternal") != InternalElement.toExternal
        ])]))

    async def build(self, **extra: Dict[InternalElement, Tuple[list, dict]]) -> "MessageChain":
        result = []
        for i in self.__root__:
            if isinstance(i, InternalElement):
                if getattr(i.__class__, "toExternal") == InternalElement.toExternal:
                    raise EntangledSuperposition("You define an object that cannot be sent: {0}".format(i.__class__.__name__))
                result.append(await run_always_await(i.toExternal(
                    *(extra[i.__class__][0] if i.__class__ in extra else []),
                    **(extra[i.__class__][1] if i.__class__ in extra else {})
                )))
            else:
                result.append(i)
        return MessageChain(__root__=tuple(result))

    def has(self, element_class: T) -> bool:
        """判断消息链中是否含有特定类型的消息元素

        Args:
            element_class (T): 需要判断的消息元素的类型, 例如 "Plain", "At", "Image" 等.

        Returns:
            bool: 判断结果
        """
        return element_class in [type(i) for i in self.__root__]

    def get(self, element_class: T) -> List[T]:
        """获取消息链中所有特定类型的消息元素

        Args:
            element_class (T): 指定的消息元素的类型, 例如 "Plain", "At", "Image" 等.

        Returns:
            List[T]: 获取到的符合要求的所有消息元素; 另: 可能是空列表([]).
        """
        return [i for i in self.__root__ if isinstance(i, element_class)]

    def asDisplay(self) -> str:
        """获取以字符串形式表示的消息链, 且趋于通常你见到的样子.

        Returns:
            str: 以字符串形式表示的消息链
        """
        return "".join(i.asDisplay() for i in self.__root__)

    @classmethod
    def join(cls, *chains: "MessageChain") -> "MessageChain":
        """拼接参数中给出的所有消息链

        Returns:
            MessageChain: 拼接结果
        """
        return cls.create(sum(chains, []))

    def plusWith(self, *chains: "MessageChain") -> "MessageChain":
        """在现有的基础上将另一消息链拼接到原来实例的尾部, 并生成, 返回新的实例.

        Returns:
            MessageChain: 拼接结果
        """
        return self.create(sum(chains, self.__root__))

    def plus(self, *chains: "MessageChain") -> NoReturn:
        """在现有的基础上将另一消息链拼接到原来实例的尾部

        Raises:
            ValueError: 原有的消息链不可变, 需要转为可变形态.

        Returns:
            NoReturn: 本方法无返回.
        """
        if self.isImmutable:
            raise ValueError("this chain is not mutable")
        for i in chains:
            self.__root__.extend(i.__root__)

    __contains__ = has
    __getitem__ = get

    def asSerializationString(self) -> str:
        return "".join([i.asSerializationString() for i in self.__root__])
    
    @classmethod
    def fromSerializationString(cls, string: str) -> "MessageChain":
        from .elements.internal import Source, Plain, AtAll, At, Face, Image, FlashImage
        code_without_argument = {
            "atall": AtAll
        }
        origin_list = list(regex.finditer(
            r"((?:\[mirai:(?P<name>[^:]+)\])|(?:\[mirai:(?P<name>[^\]]*)?:(?P<argument>.*?)?\]))|(?P<plaintext>.{1,})", string))

        result = [] # 默认为不可变型, 但这里要插入元素
        for i in origin_list:
            origin_pattern = i.groupdict()
            if origin_pattern['name'] and not origin_pattern['argument']: # 无参数型
                target_element = code_without_argument.get(origin_pattern['name'])
                if target_element:
                    result.append(target_element())
            elif origin_pattern['name'] and origin_pattern['argument']: # 有参数
                arguments = origin_pattern['argument'].split(",")
                if origin_pattern['name'] == "source":
                    result.append(Source(id=int(arguments[0]), time=arguments[1]))
                elif origin_pattern['name'] == "at":
                    result.append(At(target=int(arguments[0]), display=arguments[1]))
                elif origin_pattern['name'] == "face":
                    result.append(Face(faceId=int(arguments[0])))
                elif origin_pattern['name'] == "image":
                    result.append(Image(imageId=origin_pattern['argument']))
                elif origin_pattern['name'] == "flash":
                    result.append(FlashImage(imageId=origin_pattern['argument']))
            elif origin_pattern['plaintext']:
                result.append(Plain(origin_pattern['plaintext']))
        return cls.create(tuple(result))


# 时间片: 在时间单元内的时间(执行时间为 time_unit.base + time_shape.offset)
# 时间单元: 头尾相接的发生, 但也有例外(mouth是这样的), base是绝对时间
# 时间点: 绝对时间

from pydantic import BaseModel
from datetime import timedelta, datetime
from typing import List, Optional
from abc import abstractmethod, ABC, abstractstaticmethod

class TimeSlice(BaseModel):
    offset: timedelta

class TimeUnit(BaseModel, ABC):
    base: datetime
    children: List[TimeSlice]

    @abstractstaticmethod
    def when() -> Optional[timedelta]:
        """该抽象方法用于描述本时间单位是否开始, 如果不开始预计要在什么时候开始.

        Returns:
            Optional[timedelta]: 若为 `None` 则马上开始, 若不为则在所返回的 timedelta 所定义的时间后开始.
        """

class TimePoint(BaseModel):
    on: datetime


from typing import Union
from graia.broadcast.entities.dispatcher import BaseDispatcher
from graia.broadcast.interfaces.dispatcher import DispatcherInterface
from .exceptions import (AccountMuted, AccountNotFound, InvaildArgument,
                         InvaildAuthkey, InvaildSession, NotSupportedVersion,
                         TooLongMessage, UnauthorizedSession, UnknownTarget)
from .context import enter_context
import inspect

def applicationContextManager(func):
    def wrapper(self, *args, **kwargs):
        with enter_context(self):
            return func(self, *args, **kwargs)
    wrapper.__annotations__ = func.__annotations__
    return wrapper

def requireAuthenticated(func):
    def wrapper(self, *args, **kwargs):
        if not self.connect_info.sessionKey:
            raise InvaildSession("you must authenticate before this.")
        return func(self, *args, **kwargs)
    wrapper.__annotations__ = func.__annotations__
    return wrapper

def SinceVersion(*version: int):
    def wrapper(func):
        def inside_wrapper(self, *args, **kwargs):
            if self.connect_info.current_version and \
                self.connect_info.current_version < version:
                raise NotSupportedVersion(
                    "the current version does not support this feature: {0}".format(self.connect_info.current_version)
                )
            return func(*args, **kwargs)
        return inside_wrapper
    return wrapper

def DeprecatedSince(*version: int, action: str = "warn"):
    if action not in ["warn", "error", "ignore"]:
        raise TypeError("action must be in" + str(["warn", "error", "ignore"]))
    def wrapper(func):
        def inside_wrapper(self, *args, **kwargs):
            if self.connect_info.current_version and \
                self.connect_info.current_version > version:
                if action == "error":
                    raise NotSupportedVersion(
                        "the current version deprecated this feature: {0}".format(self.connect_info.current_version)
                    )
                elif action == "warn":
                    import warnings
                    warnings.warn(
                        "'{0}' has been deprecated since {1}, use other methods to realize your business as soon as possible!"\
                        .format(func.__qualname__, version))
            return func(*args, **kwargs)
        return inside_wrapper
    return wrapper

code_exceptions_mapping = {
    1: InvaildAuthkey,
    2: AccountNotFound,
    3: InvaildSession,
    4: UnauthorizedSession,
    5: UnknownTarget,
    6: FileNotFoundError,
    10: PermissionError,
    20: AccountMuted,
    30: TooLongMessage,
    400: InvaildArgument
}

def raise_for_return_code(code: Union[dict, int]):
    if isinstance(code, dict):
        code = code.get("code")
        exception_code = code_exceptions_mapping.get(code)
        if exception_code:
            raise exception_code
    elif isinstance(code, int):
        exception_code = code_exceptions_mapping.get(code)
        if exception_code:
            raise exception_code

def print_traceback_javay():
    stacks = inspect.stack()[1:]
    for i in stacks:
        print(f"    at [{i.filename}:{i.lineno}]")
    print("\n")

class AppMiddlewareAsDispatcher(BaseDispatcher):
    always = True

    def __init__(self, app) -> None:
        self.app = app

    def catch(self, interface: "DispatcherInterface"):
        with enter_context(self.app, interface.event):
            yield

def context_enter_auto(context):
    def wrapper1(func):
        def wrapper2(*args, **kwargs):
            with context:
                return func(*args, **kwargs)
        return wrapper2
    return wrapper1

def call_atonce(*args, **kwargs):
    def wrapper(callable_target):
        return callable_target(*args, **kwargs)
    return wrapper


from typing import Callable
from . import TimePoint, TimeSlice, TimeUnit
from datetime import datetime, timedelta
import asyncio

class DayUnit(TimeUnit):
    @staticmethod
    def when():
        now = datetime.now()
        tomorrow = (now + timedelta(days=1)).replace(minute=0, hour=0, second=0, microsecond=0)
        return tomorrow - now

class HourUnit(TimeUnit):
    @staticmethod
    def when():
        now = datetime.now()
        next_hour = (now + timedelta(hours=1)).replace(minute=0, second=0, microsecond=0)
        return next_hour - now

class MinuteUnit(TimeUnit):
    @staticmethod
    def when():
        now = datetime.now()
        next_minute = (now + timedelta(minutes=1)).replace(second=0, microsecond=0)
        return next_minute - now

class SecondSlice(TimeSlice):
    offset: timedelta = timedelta(seconds=1)

class MinuteSlice(TimeSlice):
    offset: timedelta = timedelta(minutes=1)

class HourSlice(TimeSlice):
    offset: timedelta = timedelta(hours=1)

class DaySlice(TimeSlice):
    offset: timedelta = timedelta(days=1)

class MonthSlice(TimeSlice):
    @property
    def offset(self) -> timedelta:
        now_month = datetime.now().month
        if now_month in (1, 3, 5, 7, 8, 10, 12):
            return timedelta(days=30)
        else:
            return timedelta(days=31)

class SchedulerControl:
    def __init__(self, loop: asyncio.AbstractEventLoop) -> None:
        self.loop = loop
    
    def schedule_callable(self, callable: Callable, unit: TimeUnit, *args, **kwargs) -> asyncio.Task:
        """使用 asyncio 提供的各式 API, 将给出的 Callable 对象作为调度对象, 并根据所给出的时间单位作出决策.

        Args:
            callable (Callable): 任意调度对象
            unit (TimeUnit): 时间单位

        Returns:
            asyncio.Task: undefined.
        """
        pass


from pydantic import BaseModel, AnyHttpUrl
from typing import Tuple
import yarl

class Session(BaseModel):
    """用于描述与上游接口会话, 并存储会话状态的实体类.
    
    Attributes:
        host (AnyHttpUrl): `mirai-api-http` 服务所在的根接口地址
        authKey (str): 在 `mirai-api-http` 配置流程中定义, 需为相同的值以通过安全验证.
        account (int): 应用所使用账号的整数 ID.
        websocket (bool): 是否使用 Websocket 方式获取事件, 若为 `False` 则使用 HTTP 短轮询方式获取事件, 性能较低.
        sessionKey (str, optional): 会话标识, 即会话中用于进行操作的唯一认证凭证, 需要经过 `activeSession` 后才可用.
        current_version (Tuple[int, int, int], optional): 上游服务的版本, 暂时没有自动获取.
    """

    host: AnyHttpUrl
    authKey: str
    account: int
    sessionKey: str = None
    current_version: Tuple[int, int, int] = None
    websocket: bool = True

    @classmethod
    def fromUrl(cls, url: str) -> "Session":
        """不再推荐的声明连接方式的方法, 用于兼容 v3 中的 url 声明连接信息的方法.

        Args:
            url (str): `schema` 为 `mirai` 的链接, 具体请参考 python-mirai(v3) 的文档.
        
        Returns:
            Session: 从给出的 url 生成的 `Session` 实例.
        """

        yarl_url = yarl.URL(url)
        return cls(
            host=str(yarl.URL.build(
                scheme="http",
                host=yarl_url.host,
                port=yarl_url.port,
            )),
            authKey=yarl_url.query["authKey"],
            account=yarl_url.query["qq"],
            websocket=yarl_url.path == "/ws"
        )

    class Config:
        allow_mutation = True